/********************************************************/
/*							*/
/*	Daemon which handle backup			*/
/*							*/
/********************************************************/
#include	<stdlib.h>
#include	<getopt.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<signal.h>
#include	<wait.h>

#include	"modbck.h"
#include	"gescnf.h"
#include	"subrou.h"

#define	READER		"reader"
#define	MARKER		"marker"
#define	BLKSIZE		32768

#define	SENDMSG		"sendmsg.sh"

#define	SIGPROBE	0	//send signal 0 to a process

typedef enum {		/*application type		*/
	backd,
	marker,
	reader
	}WHATTYPE;

typedef	enum {		/*status type			*/
	s_ok,		/*backup successfully terminated*/
	s_notape,	/*no tape avaliable		*/
	s_notinserted,	/*tape not in tape reader.	*/
	s_tapein,	/*tape ready and inserted	*/
	s_noslot,	/*no slot found on tape		*/
	s_toobig,	/*data volume too big for tape	*/
	s_wrongtape,	/*tape ready and inserted	*/
	s_unknown	/*unknown error			*/
	}STATYPE;

static int hangup;	/*received the HANGUP signal	*/
/*

*/
/********************************************************/
/*							*/
/*	Daemon usage flags				*/
/*							*/
/********************************************************/
static void usage()

{
(void) fprintf(stderr,"%s-%s\n",APPNAME,getvers());
(void) fprintf(stderr,"\tusage:  %s [-b base] [-d #] [-f] [-n] [-v] [-h] "
		      "[-V]\n",APPNAME);
(void) fprintf(stderr,"\tusage:  %s [-d #] [-v] tapeid [tapeid]"
		      "\n",MARKER);
(void) fprintf(stderr,"\tusage:  %s [-d #] [-v] tapedev [slot #]\n",
		      READER);
(void) fprintf(stderr,"\t%s\t\t: Used to do write backup tape\n",APPNAME);
(void) fprintf(stderr,"\t\t-b newbase\t: operational base\n");
(void) fprintf(stderr,"\t\t-d #\t\t: debug level\n");
(void) fprintf(stderr,"\t\t-e\t\t: do not eject tape when backup completed\n");
(void) fprintf(stderr,"\t\t-f\t\t: Work in foregrounde mode (debug purpose)\n");
(void) fprintf(stderr,"\t\t-h\t\t: display usage\n");
(void) fprintf(stderr,"\t\t-n\t\t: start backup at once\n");
(void) fprintf(stderr,"\t\t-v\t\t: get version \n");
(void) fprintf(stderr,"\t\t-V\t\t: verbose mode\n");
(void) fprintf(stderr,"\t%s\t\t: Used to label tape\n",MARKER);
(void) fprintf(stderr,"\t\ttapeid\t\t: Uniq ID for the tape media\n");
(void) fprintf(stderr,"\t\t\t\t: label as many tapes as there is TAPEID\n");
(void) fprintf(stderr,"\t%s\t\t: Used to read tape\n",READER);
(void) fprintf(stderr,"\t\ttapedev\t\t: physical tape reader\n");
(void) fprintf(stderr,"\t\tslot #\t\t: which tape slot you want to extract from the tape\n");
(void) fprintf(stderr,"\t\t\t\t: (if not specified, extract the last one)\n");
}
/*

*/
/********************************************************/
/*							*/
/*	Sub routine to trappe signal event		*/
/*							*/
/********************************************************/
static void sigalrm(int sig)

{
if (sig!=SIGALRM)
  (void) rou_alert(9,"backdc:sigalrm signal <%s> received",strsignal(sig));
switch (sig)
  {
  case SIGCHLD		:
    while (waitpid(-1,(int *)0,WNOHANG)>0);
    break;
  case SIGQUIT		:
  case SIGTERM		:
  case SIGHUP		:
    hangup=true;
    break;
  case SIGALRM		:
    break;
  default        	:
    (void) rou_alert(0,"Signal [%d] received",sig);
    break;
  }
(void) signal(sig,sigalrm);
}
/*

*/
/********************************************************/
/*							*/
/*	Init interrupts					*/
/*							*/
/********************************************************/
static void initinterrupts()

{
(void) signal(SIGCHLD,sigalrm);
(void) signal(SIGHUP,sigalrm);
(void) signal(SIGQUIT,sigalrm);
(void) signal(SIGTERM,sigalrm);
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to find out what is the backd 	*/
/*	inpersonnation.					*/
/*							*/
/********************************************************/
static WHATTYPE whatami(char *myname)

{
char *name;

if ((name=strrchr(myname,'/'))!=(char *)0)
  name++;
else
  name=myname;
if (!strcmp(MARKER,name))
  return marker;
if (!strcmp(READER,name))
  return reader;
return backd;
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to send an acknowledge to backup	*/
/*	administrator.					*/
/*							*/
/********************************************************/
static void acknowledge(BCKTYPE *bckinfo,STATYPE status)

{
static char *vstat[]={"S_OK",
		      "S_NOTAPE",
		      "S_MISSING",
		      "S_TAPEIN",
		      "S_NOSLOT",
		      "S_TOOBIG",
		      "S_WRONGTAPE",
		      "S_UNKNOWN"
		     };

FILE *file;
char pathname[300];

(void) sprintf(pathname,"%s/%s/%s \"%s\" \"%s\" \"%s\"",
			 base,SHELL,SENDMSG,vstat[status],
			 bckinfo->tnum,
		         localedate(bckinfo->stamp));
if ((file=popen(pathname,"w"))!=(FILE *)0) {
  if (status==s_ok)
    (void) dumpbckinfofile(bckinfo,file);
  (void) pclose(file);
  }
else
  (void) rou_alert(0,"backd.c,acknowledge: Unable to execute '%s'",pathname);
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to put the pidlock to zero on the	*/
/*	tape entry.					*/
/*							*/
/********************************************************/
static void lockzero(TAPETYPE **tinfo,BCKTYPE *bckinfo)

{
if (tinfo!=(TAPETYPE **)0) {
  int i;

  for (i=0;tinfo[i]!=(TAPETYPE *)0;i++) {
    if (tinfo[i]->tapeid!=(char *)0) {
      if (strcmp(tinfo[i]->tapeid,bckinfo->tnum)==0) {
        tinfo[i]->pidlock=(pid_t)0;
        tinfo[i]->lastused=bckinfo->lastused;
        tinfo[i]->cycled=bckinfo->num;
        tinfo[i]->frozen=bckinfo->tobkept;
        break;
        }
      }
    }
  }
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to prepare a backup to do		*/
/*							*/
/********************************************************/
static BCKTYPE *prepbackup(TAPETYPE **tinfo)

{
BCKTYPE *bckinf;

bckinf=(BCKTYPE *)0;
if (tinfo!=(TAPETYPE **)0) {
  IDTYPE *idinfo;

  if ((bckinf=getbackup())!=(BCKTYPE *)0) {
    if ((idinfo=getidinfo(tinfo,bckinf->tape,bckinf->tobkept))!=(IDTYPE *)0) {
      bckinf->blksize=idinfo->blksize;
      bckinf->block=idinfo->blocks;
      bckinf->num=idinfo->cycled;
      bckinf->lastused=idinfo->lastused;
      (void) strcpy(bckinf->tdev,idinfo->tdev);
      (void) strcpy(bckinf->tnum,idinfo->tapeid);
      }
    idinfo=cleanidinfo(idinfo);
    }
  }
return bckinf;
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to extract a specific slot from the	*/
/*	tape.						*/
/*							*/
/********************************************************/
static void extracting(DEVTYPE *devinfo,int argc,char *argv[])

{
#define	OPEP	"backd.c:extracting,"

THANDLE *thandle;
int phase;
int proceed;

thandle=(THANDLE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening device		*/
      if ((thandle=dev_open(devinfo->tdev,devinfo->blksize,O_RDONLY))==(THANDLE *)0) {
        (void) fprintf(stderr,"Unable to open device '%s'\n",devinfo->tdev);
	phase=999;	/*no need to go further		*/
	}
      break;
    case 1	:	/*positionning cursor		*/
      {
      int slot;

      slot=ENDFILE;
      if (argc>0) 
        slot=atoi(argv[0]);
      if (dev_move(thandle,slot)<0) {
        (void) fprintf(stderr,"Unable to set device '%s' cursor\n",devinfo->tdev);
	phase=999;	/*no need to go further		*/
	}
      }
      break;
    case 2	:	/*extracting data		*/
      {
      char *buffer;
      siz64_t lu;

      buffer=(char *)calloc(1,devinfo->blksize);
      while ((lu=dev_read(thandle,buffer,devinfo->blksize))>0) 
	(void) fwrite(buffer,1,lu,stdout);
      (void) free(buffer);
      }
      break;
    default	:	/*SAFE Guard			*/
      (void) dev_close(thandle,false);
      proceed=false;
      break;
    }
  phase++;
  }
#undef	OPEP
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to proceed with actual backup	*/
/*	return with the waiting time to do		*/
/*							*/
/********************************************************/
static int dobackd(int atonce,int eject)

{
#define	OPEP	"backd.c:dobackd,"

TAPETYPE **tinfo;
BCKTYPE *bckinfo;
int ready;

tinfo=loadtapelist();
if ((bckinfo=prepbackup(tinfo))==(BCKTYPE *)0) {
  (void) rou_alert(0,"No backup to do!, exiting!");
  tinfo=cleantapelist(tinfo);
  return 3600;
  }
if (strcasecmp(bckinfo->tnum,"none")==0) {
  (void) acknowledge(bckinfo,s_notape);
  return (12*3600);
  }
ready=false;
while (true) {
  int sleepy;

  if (systime()>bckinfo->stamp)
    break;
  switch (checktape(bckinfo->tdev,bckinfo->tnum,bckinfo->blksize,true,1)) {
    case (a_ok)		:
      if (ready==false) {
        (void) acknowledge(bckinfo,s_tapein);
	ready=true;
        }
      break;
    case (tmissing)	:
      (void) acknowledge(bckinfo,s_notinserted);
      ready=false;
      atonce=false;
      break;
    case (noslot)	:
      (void) acknowledge(bckinfo,s_noslot);
      ready=false;
    case (toobig)	:
      (void) acknowledge(bckinfo,s_toobig);
      ready=false;
    case (wrongtape)	:
      (void) acknowledge(bckinfo,s_wrongtape);
      ready=false;
      break;
    } 
  sleepy=(bckinfo->stamp-systime())/2;
  if ((atonce==true)||(sleepy<0))
    break;
  if (sleepy>(24*60*60))
    sleepy=24*60*60;
  if (sleepy<60*10)
    sleepy=60*10;
  (void) rou_alert(0,"Going to sleep for '%d' sec (ready status='%d')",sleepy,ready);
  (void) sleep(sleepy);
  if (hangup==true) {
    ready=false;
    break;		//exiting from waiting loop
    }
  }
(void) rou_alert(0,"Going to do backup now on tape <%s> (ready status='%d')",
		    bckinfo->tnum,ready);
if (ready==true) {
  switch (dobackup(bckinfo,eject,true)) {
    case (a_ok)		:
      bckinfo->lastused=normdate(bckinfo->stamp);
      (void) acknowledge(bckinfo,s_ok);
      break;
    case (tmissing)	:
      (void) acknowledge(bckinfo,s_notinserted);
      break;
    case (wrongtape)	:
      (void) acknowledge(bckinfo,s_wrongtape);
      break;
    default		:
      (void) acknowledge(bckinfo,s_unknown);
      break;
    }
  }
(void) lockzero(tinfo,bckinfo);
(void) dumptapelist(tinfo);
bckinfo=cleanbckinfo(bckinfo);
tinfo=cleantapelist(tinfo);
return 10;
#undef	OPEP
}
/*

*/
/********************************************************/
/*							*/
/*	Subroutine to read the tape.			*/
/*							*/
/********************************************************/
static void doreading(int argc,char *argv[])

{
if (argc==0)
  (void) usage();
else {
  DEVTYPE **devinfo;
  int found;
  int i;

  found=false;
  if ((devinfo=loaddevlist())==(DEVTYPE **)0) {
    (void) fprintf(stderr,"Unable to load device list\n");
    return;
    }
  for (i=0;devinfo[i]!=(DEVTYPE *)0;i++) {
    if (strcmp(devinfo[i]->tdev,argv[0])==0) {
      (void) extracting(devinfo[i],argc-1,&(argv[1]));
      found=true;
      break;
      }
    }
  devinfo=cleandevlist(devinfo);
  if (found==false)
    (void) fprintf(stderr,"Unable to find device '%s' in the device list\n",argv[0]);
  }
}
/*

*/
/********************************************************/
/*							*/
/*	Main routine					*/
/*							*/
/********************************************************/
int main(int argc,char *argv[])

{
char c;
int atonce;
int eject;

atonce=false;
eject=true;
foreground=false;
while ((c=getopt(argc,argv,"Vb:d:efhnv"))!=EOF) {
  switch ((int)c) {
    case (int) 'V':
      verbose++;
      break;
    case (int) 'b':
      base=optarg;
      break;
    case (int) 'd':
      debug=atoi(optarg);
      break;
    case (int) 'e':
      eject=false;
      break;
    case (int) 'f':
      foreground=true;
      break;
    case (int) 'n':
      atonce=true;
      break;
    case (int) 'v':
      (void) fprintf(stdout,"%s\n",getversion());
      (void) exit(0);
      break;
    case (int) 'h':
    default       :
      (void) usage();
      (void) exit(-1);
      break;
    }
  }
(void) initinterrupts();
switch (whatami(argv[0])) {
  case backd	:
    if (foreground==false) {
      verbose=false;
      if (rou_divedivedive(foreground)!=0) {
        (void) exit(0);	//we are now in daemon mode
	}
      }
    (void) openlog(appname,LOG_PID,LOG_DAEMON);
    if (rou_locking((const char *)0,LCK_LOCK,2)==false) {
      (void) exit(0);	//trouble trouble
      }
    do {
      pid_t subback;

      subback=rou_divedivedive(foreground);
      switch(subback) {
	case -1	:	//trouble, trouble
	  (void) rou_alert(0,"Unable to start '%s' in background (error=<%s>)",
		 	      appname,strerror(errno));
          break;
	case 0	:
          (void) dobackd(atonce,eject);
          break;
	default	:	// waiting sub backd program
          (void) rou_alert(0,"Started %s-%s in background mode (pid='%05d')\n",
		      	      appname,getvers(),subback);
	  while (rou_checkprocess(subback)==true) {
	    (void) sleep(30);
	    if (hangup==true) {
	     (void) kill(subback,SIGTERM);
	     (void) sleep(1);
	     }
	    }
          (void) rou_alert(0,"%s-%s background process exited (pid='%05d')\n",
			      appname,getvers(),subback);
          break;
	}
      }
    while ((atonce==false)&&(hangup==false));
    (void) rou_locking((const char *)0,LCK_UNLOCK,2);
    (void) closelog();
    break;
  case marker	:
    verbose=true;
    (void) dolabel(argc-optind,&(argv[optind]));
    break;
  case reader	:
    verbose=true;
    (void) doreading(argc-optind,&(argv[optind]));
    break;
  }
return 0;
}
